@@ -3,34 +3,22 @@ import sys
 import logging
 import click
 import pkg_resources
-import inspect
+from datetime import datetime
 from uuid import uuid4
 from flask import Response, redirect
 from flask_compress import Compress
 from flask_cors import CORS
-from pymacaron_core.swagger.apipool import ApiPool
-from pymacaron_core.models import get_model
-import pymacaron.models
 from pymacaron.apiloader import load_api_models_and_endpoints
 from pymacaron.log import set_level, pymlogger
-from pymacaron.crash import set_error_reporter, generate_crash_handler_decorator
-from pymacaron.exceptions import format_error
 from pymacaron.config import get_config
 from pymacaron.monitor import monitor_init
+from pymacaron.crash import set_error_reporter
 from pymacaron.api import add_ping_hook
 
 
 log = pymlogger(__name__)
 
 
-# TODO: deprecate
-def _get_model_factory(model_name):
-    # Using dynamic method creation to localize model_name
-    def factory(**kwargs):
-        return get_model(model_name)(**kwargs)
-    return factory
-
-
 class modelpool():
     """The modelpool of an api is a class whose attributes are all the Pymacaron
     model classes declared in that api
@@ -42,16 +30,19 @@ class modelpool():
     def __getattr__(self, model_name):
         raise Exception(f'Either {self.api_name}.yaml has not been loaded or it does not define object {model_name}')
 
-    def json_to_model(self, model_name, j, keep_datetime=None):
+    def json_to_model(self, model_name, j, keep_datetime=None, prune_none=True):
         """Given a model name and json dict, return an instantiated pymacaron object"""
         # TODO: keep_datetime in this method should be deprecated...
         assert keep_datetime is not False, "Support for keep_datetime=False not implemented"
-        return getattr(self, model_name).from_json(j)
+        return getattr(self, model_name).from_json(j, prune_none=prune_none)
 
 
 class apipool():
     """The apipool contains the modelpools of all loaded apis"""
 
+    # api_name: api_path
+    __api_paths = {}
+
     @classmethod
     def add_model(cls, api_name, model_name, model_class):
         """Register a model class defined in an api"""
@@ -69,22 +60,115 @@ class apipool():
         assert hasattr(apipool, api_name), f"Api {api_name} is not loaded in apipool"
         return getattr(apipool, api_name)
 
-
-class apispecs():
-    """Keep track of the path of the openapi spec file of every loaded api"""
-
-    __api_name_to_path = {}
+    @classmethod
+    def get_api_names(cls):
+        """Return the names of loaded apis"""
+        return sorted(apipool.__api_paths.keys())
 
     @classmethod
-    def register_api_path(cls, api_name, api_path):
-        """Remember where to find the openapi file of this api"""
-        apispecs.__api_name_to_path[api_name] = api_path
+    def load_swagger(cls, api_name, api_path, dest_dir=None, create_endpoints=False, force=False):
+        """Load a swagger/openapi specification into pymacaron: generate its model
+        classes (declared with pydantic), and optionally generate the Flask api
+        endpoints binding endpoint methods to routes.
+
+        Syntax:
+            apipool.load_swagger('ping', '../apis/ping.yaml')
+
+        api_name : str
+            Name of the api, used to access api models.
+        api_path : str
+            Path of the swagger file of the api.
+        dest_dir: str
+            Optional. Path to a directory under which to write the generated
+            '<api_name>_models.py' and '<api_name>_app.py' files. Defaults to
+            the same directory as the swagger file.
+        create_endpoints: bool
+            Optional. Set to false to only generate model declarations, and not
+            endpoint declarations. Defaults to true.
+        force: bool
+            Optional. Force regenerating the model and endpoint code even if
+            the code files are up to date with the swagger file. Defaults to
+            false.
+
+        """
+
+        app_pkg = load_api_models_and_endpoints(
+            api_name=api_name,
+            api_path=api_path,
+            dest_dir=dest_dir,
+            create_endpoints=create_endpoints,
+            force=force,
+        )
+
+        # Remember where this api's swagger is located
+        apipool.__api_paths[api_name] = api_path
+
+        return app_pkg
 
     @classmethod
-    def get_api_path(cls, api_name):
-        """Get the path of the openapi file of this api, or None if that api has not been loaded"""
-        return apispecs.__api_name_to_path.get(api_name, None)
+    def publish_apis(cls, app, path='doc'):
+        """Add routes to the Flask app to publish all loaded swagger files under the
+        paths doc/<api_name>.yaml and doc/<api_name>
+        """
 
+        if not apipool.__api_paths:
+            raise Exception("You must call .load_apis() before .publish_apis()")
+
+        # Infer the live host url from pym-config.yaml
+        proto = 'http'
+        if hasattr(get_config(), 'aws_cert_arn'):
+            proto = 'https'
+
+        live_host = "%s://%s" % (proto, get_config().live_host)
+
+        # Allow cross-origin calls
+        CORS(app, resources={r"/%s/*" % path: {"origins": "*"}})
+
+        # Add routes to serve api specs and redirect to petstore ui for each one
+        for api_name, api_path in apipool.__api_paths.items():
+
+            api_filename = os.path.basename(api_path)
+            log.info("Publishing api %s at /%s/%s" % (api_name, path, api_name))
+
+            def redirect_to_petstore(live_host, api_filename):
+                def f():
+                    url = 'http://petstore.swagger.io/?url=%s/%s/%s' % (live_host, path, api_filename)
+                    log.info("Redirecting to %s" % url)
+                    return redirect(url, code=302)
+                return f
+
+            def serve_api_spec(api_path):
+                def f():
+                    with open(api_path, 'r') as f:
+                        spec = f.read()
+                        log.info("Serving %s" % api_path)
+                        return Response(spec, mimetype='text/plain')
+                return f
+
+            # Publish yaml file and a redirect link to petstore's swagger UI to visualize that file
+            app.add_url_rule('/%s/%s' % (path, api_name), str(uuid4()), redirect_to_petstore(live_host, api_filename))
+            app.add_url_rule('/%s/%s' % (path, api_filename), str(uuid4()), serve_api_spec(api_path))
+
+
+class jsonencoders:
+
+    # Default encoder when serializing datetime to Flask response
+    __json_encoders = {
+        datetime: lambda d: d.isoformat(),
+    }
+
+    @classmethod
+    def set_json_encoders(cls, encoders):
+        assert type(encoders) is dict, "json_encoders must be a dict of type/callable"
+        assert datetime in encoders, "json_encoders must define an encoder for datetime"
+        assert callable(encoders[datetime]), "json_encoders[datetime] must be a callable"
+        cls.__json_encoders = encoders
+
+
+    @classmethod
+    def get_datetime_encoder(cls):
+        assert datetime in jsonencoders.__json_encoders
+        return jsonencoders.__json_encoders[datetime]
 
 
 def get_port():
@@ -110,17 +194,26 @@ def get_port():
 class API(object):
 
 
-    def __init__(self, app, host='localhost', port=None, debug=False, log_level=logging.DEBUG, formats=None, timeout=20, error_reporter=None, default_user_id=None, error_callback=format_error, error_decorator=None, ping_hook=[]):
+    def __init__(self, app, host='localhost', port=None, debug=False, log_level=logging.DEBUG, json_encoders=None, error_reporter=None, error_callback=None, default_user_id=None, ping_hook=[]):
         """
 
         Configure the Pymacaron microservice prior to starting it. Arguments:
 
-        - app: the flask app
-        - port: (optional) the http port to listen on (defaults to 80)
-        - debug: (optional) whether to run with flask's debug mode (defaults to False)
-        - error_reporter: (optional) a callback to call when catching exceptions, for custom reporting to slack, email or whatever
-        - log_level: (optional) the microservice's log level (defaults to logging.DEBUG)
-        - ping_hook: (optional) a function to call each time Amazon calls the ping endpoint, which happens every few seconds
+        app : the flask app
+
+        port : (optional) the http port to listen on (defaults to 80)
+
+        debug : (optional) whether to run with flask's debug mode (defaults to False)
+
+        error_callback : (optional) takes the exception caught while executing an endpoint and return its json representation in the microservice's own error format
+
+        error_reporter : (optional) takes the exception caught in an endpoint together with a title and long text trace, and reports them wherever is relevant
+
+        log_level : (optional) the microservice's log level (defaults to logging.DEBUG)
+
+        ping_hook : (optional) a function to call each time Amazon calls the ping endpoint, which happens every few seconds
+
+        json_encoders: (optional) custom pydantic json encoders, to use when serializing pymacaron models to json
 
         """
         assert app
@@ -130,11 +223,10 @@ class API(object):
         self.port = port
         self.host = host
         self.debug = debug
-        self.formats = formats
-        self.timeout = timeout
         self.error_callback = error_callback
-        self.error_decorator = error_decorator
+        self.error_reporter = error_reporter
         self.ping_hook = ping_hook
+        self.app_pkgs = []
 
         if not port:
             self.port = get_port()
@@ -144,40 +236,36 @@ class API(object):
 
         set_level(log_level)
 
-        if error_reporter:
-            set_error_reporter(error_reporter)
+        if json_encoders:
+            log.info(f"Using custom json encoder when serializing Flask response: {json_encoders}")
+            jsonencoders.set_json_encoders(json_encoders)
 
         log.info("Initialized API (%s:%s) (Flask debug:%s)" % (host, port, debug))
 
 
-    def load_default_apis(self):
-        # Load models from default builtin apis
-        for name in ['ping', 'crash']:
+    def publish_apis(self, path='doc'):
+        """Publish all loaded apis on under the uri /<path>/<api-name>, by
+        redirecting to http://petstore.swagger.io/
+        """
+        apipool.publish_apis(self.app, path=path)
+
+
+    def load_builtin_apis(self, names=['ping']):
+        """Load some or all of the builtin apis 'ping' and 'crash'"""
+        for name in names:
             yaml_path = pkg_resources.resource_filename(__name__, 'pymacaron/%s.yaml' % name)
             if not os.path.isfile(yaml_path):
                 yaml_path = os.path.join(os.path.dirname(sys.modules[__name__].__file__), '%s.yaml' % name)
-            load_api_models_and_endpoints(
-                api_name=name,
-                api_path=yaml_path,
-                dest_path=get_config().apis_path,
-                load_models=True,
-                load_endpoints=True,
+            app_pkg = apipool.load_swagger(
+                name,
+                yaml_path,
+                dest_dir=get_config().apis_path,
+                create_endpoints=True,
             )
+            if app_pkg:
+                self.app_pkgs.append(app_pkg)
 
 
-    # TODO: deprecate
-    def _load_model_aliases(self, api):
-        """Load all PyMacaronModels generated for a given api into the namespace
-        of pymacaron.models so a user may write 'from pymacaron.models import <SomeModel>'
-        """
-
-        for model_name in dir(api.model):
-            model_class = getattr(api.model, model_name)
-            if inspect.isclass(model_class) and 'to_json' in dir(model_class):
-                setattr(pymacaron.models, model_name, _get_model_factory(model_name))
-
-
-    # TODO: deprecate
     def load_clients(self, path=None, apis=[]):
         """Generate client libraries for the given apis, without starting an
         api server"""
@@ -195,37 +283,39 @@ class API(object):
             api_path = os.path.join(path, '%s.yaml' % api_name)
             if not os.path.isfile(api_path):
                 raise Exception("Cannot find swagger specification at %s" % api_path)
-            log.info("Loading api %s from %s" % (api_name, api_path))
-            api = ApiPool.add(
+            apipool.load_swagger(
                 api_name,
-                yaml_path=api_path,
-                timeout=self.timeout,
-                error_callback=self.error_callback,
-                formats=self.formats,
-                do_persist=False,
-                local=False,
+                api_path,
+                dest_dir=path,
+                create_endpoints=False,
             )
-            self._load_model_aliases(api)
 
         return self
 
 
-    def load_apis(self, path, ignore=[], include_crash_api=False):
-        """Load all swagger files found at the given path, except those whose
-        names are in the 'ignore' list"""
+    def load_apis(self, path=None, ignore=[], only_models=[], force=False):
+        """Find all swagger files under the given path. Ignore those whose name is in
+        the ignore list. Generate and load models for all others. Generate
+        Flask app code for all except those in ignore and only_models list.
 
-        path = get_config().apis_path
+        """
+
+        if not path:
+            path = get_config().apis_path
 
         if type(ignore) is not list:
             raise Exception("'ignore' should be a list of api names")
 
+        if type(only_models) is not list:
+            raise Exception("'ignore' should be a list of api names")
+
         # Always ignore pym-config.yaml
         ignore.append('pym-config')
 
         # Find all swagger apis under 'path'
         apis = {}
 
-        log.debug("Searching path %s" % path)
+        log.debug("Searching swagger files under path %s" % path)
         for root, dirs, files in os.walk(path):
             for f in files:
                 if f.startswith('.#') or f.startswith('#'):
@@ -240,92 +330,31 @@ class API(object):
                     apis[api_name] = os.path.join(path, f)
                     log.debug("Found api %s in %s" % (api_name, f))
 
-        # And add pymacaron's default ping and crash apis
-        for name in ['ping', 'crash']:
-            yaml_path = pkg_resources.resource_filename(__name__, 'pymacaron/%s.yaml' % name)
-            if not os.path.isfile(yaml_path):
-                yaml_path = os.path.join(os.path.dirname(sys.modules[__name__].__file__), '%s.yaml' % name)
-            apis[name] = yaml_path
-
-        if not include_crash_api:
-            del apis['crash']
-
-        # Save found apis
-        self.path_apis = path
-        self.apis = apis
+        # Now generate the model and app code for all these files
+        for api_name, api_path in apis.items():
+            app_pkg = apipool.load_swagger(
+                api_name,
+                api_path,
+                dest_dir=os.path.dirname(api_path),
+                create_endpoints=False if api_name in only_models else True,
+                force=force,
+                # TODO: custom formats
+                # formats=self.formats,
+            )
+            if app_pkg:
+                self.app_pkgs.append(app_pkg)
 
         return self
 
 
-    def publish_apis(self, path='doc'):
-        """Publish all loaded apis on under the uri /<path>/<api-name>, by
-        redirecting to http://petstore.swagger.io/
-        """
-
-        assert path
-
-        if not self.apis:
-            raise Exception("You must call .load_apis() before .publish_apis()")
-
-        # Infer the live host url from pym-config.yaml
-        proto = 'http'
-        if hasattr(get_config(), 'aws_cert_arn'):
-            proto = 'https'
-
-        live_host = "%s://%s" % (proto, get_config().live_host)
-
-        # Allow cross-origin calls
-        CORS(self.app, resources={r"/%s/*" % path: {"origins": "*"}})
-
-        # Add routes to serve api specs and redirect to petstore ui for each one
-        for api_name, api_path in self.apis.items():
-
-            api_filename = os.path.basename(api_path)
-            log.info("Publishing api %s at /%s/%s" % (api_name, path, api_name))
-
-            def redirect_to_petstore(live_host, api_filename):
-                def f():
-                    url = 'http://petstore.swagger.io/?url=%s/%s/%s' % (live_host, path, api_filename)
-                    log.info("Redirecting to %s" % url)
-                    return redirect(url, code=302)
-                return f
-
-            def serve_api_spec(api_path):
-                def f():
-                    with open(api_path, 'r') as f:
-                        spec = f.read()
-                        log.info("Serving %s" % api_path)
-                        return Response(spec, mimetype='text/plain')
-                return f
-
-            self.app.add_url_rule('/%s/%s' % (path, api_name), str(uuid4()), redirect_to_petstore(live_host, api_filename))
-            self.app.add_url_rule('/%s/%s' % (path, api_filename), str(uuid4()), serve_api_spec(api_path))
-
-        return self
-
-
-    def start(self, serve=[]):
+    def start(self):
         """Load all apis, either as local apis served by the flask app, or as
         remote apis to be called from whithin the app's endpoints, then start
         the app server"""
 
-        # Check arguments
-        if type(serve) is str:
-            serve = [serve]
-        elif type(serve) is list:
-            pass
-        else:
-            raise Exception("'serve' should be an api name or a list of api names")
+        self.app.secret_key = os.urandom(24)
 
-        if len(serve) == 0:
-            raise Exception("You must specify at least one api to serve")
-
-        for api_name in serve:
-            if api_name not in self.apis:
-                raise Exception("Can't find %s.yaml (swagger file) in the api directory %s" % (api_name, self.path_apis))
-
-        app = self.app
-        app.secret_key = os.urandom(24)
+        set_error_reporter(self.error_reporter)
 
         # Initialize JWT config
         conf = get_config()
@@ -336,66 +365,25 @@ class API(object):
                 conf.jwt_secret[0:8],
             ))
 
-        # Always serve the ping api
-        serve.append('ping')
-
         # Add ping hooks if any
         if self.ping_hook:
             add_ping_hook(self.ping_hook)
 
+        self.load_builtin_apis()
+
         # Let's compress returned data when possible
         compress = Compress()
-        compress.init_app(app)
+        compress.init_app(self.app)
 
-        # All apis that are not served locally are not persistent
-        not_persistent = []
-        for api_name in self.apis.keys():
-            if api_name in serve:
-                pass
-            else:
-                not_persistent.append(api_name)
-
-        # Now load those apis into the ApiPool
-        for api_name, api_path in self.apis.items():
-
-            host = None
-            port = None
-
-            if api_name in serve:
-                # We are serving this api locally: override the host:port specified in the swagger spec
-                host = self.host
-                port = self.port
-
-            do_persist = True if api_name not in not_persistent else False
-            local = True if api_name in serve else False
-
-            log.info("Loading api %s from %s (persist: %s)" % (api_name, api_path, do_persist))
-            api = ApiPool.add(
-                api_name,
-                yaml_path=api_path,
-                timeout=self.timeout,
+        # Now execute Flask code declaring API routes
+        for app_pkg in self.app_pkgs:
+            app_pkg.load_endpoints(
+                app=self.app,
                 error_callback=self.error_callback,
-                formats=self.formats,
-                do_persist=do_persist,
-                host=host,
-                port=port,
-                local=local,
+                # formats=self.formats,
+                # local=False,
             )
 
-            self._load_model_aliases(api)
-
-        # Make sure schema objects from different APIs don't conflict with each other
-        ApiPool.merge()
-
-        # Now spawn flask routes for all endpoints
-        for api_name in self.apis.keys():
-            if api_name in serve:
-                log.info("Spawning api %s" % api_name)
-                api = getattr(ApiPool, api_name)
-                # Spawn api and wrap every endpoint in a crash handler that
-                # catches replies and reports errors
-                api.spawn_api(app, decorator=generate_crash_handler_decorator(self.error_decorator))
-
         log.debug("Argv is [%s]" % '  '.join(sys.argv))
         if 'celery' in sys.argv[0].lower():
             # This code is loading in a celery worker - Don't start the actual flask app.
@@ -403,7 +391,7 @@ class API(object):
             return
 
         # Initialize monitoring, if any is defined
-        monitor_init(app=app, config=conf)
+        monitor_init(app=self.app, config=conf)
 
         if os.path.basename(sys.argv[0]) == 'gunicorn':
             # Gunicorn takes care of spawning workers
@@ -411,9 +399,9 @@ class API(object):
             return
 
         # Debug mode is the default when not running via gunicorn
-        app.debug = self.debug
+        self.app.debug = self.debug
 
-        app.run(host='0.0.0.0', port=self.port)
+        self.app.run(host='0.0.0.0', port=self.port)
 
 #
 # Generic code to start server, from command line or via gunicorn
